; testFPM.asm from github cosmacl FPGA project
; https://github.com/wel97459/FPGACosmacELF
; additional fiddling and commentary by Herb Johnson Jan 9 2021
 
	incl	"1802reg.asm"

fpm_add 	equ	0400H
fpm_sub 	equ	02e9H ;correction from Dr Dobbs #44 letter
fpm_mult 	equ	02b7H
fpm_div 	equ	034cH
fpm_02AC 	equ	03c3H
fpm_02OP 	equ	03d9H
fpm_sqroot equ	0773H
fpm_D2AC 	equ	0500H
fpm_AC2D 	equ	0598H
fpm_V2VC 	equ	06daH
fpm_V2OP 	equ	0705H
fpm_AC2V 	equ	072cH
fpm_load 	equ	0754H

D_sign	equ	0114H ;start of decimal register
D_exp1	equ	011EH ;end of decimal register

;set up program and RAM/stack addresses

ramEnd	equ	00ffH ; stack pointer start
    org  0000H

;caution, these routines may assume page 0 addressing!
init:      ; setup the new progam counter and SCRT regs in page 0
    ghi r0
    phi r3
    phi r6        ;sets page to 00H
    ldi low main  ;R3 program counter to main
    plo r3
    plo r6

; setup call
    ghi r0
    phi r4        ;sets page to 00H
    ldi low call
    plo r4

; setup return
    ghi r0
    phi r5        ;sets page to 00H
    ldi low return
    plo r5

; setup stack at end of page 0 memory
    ldi high ramEnd
    phi r2
    ldi low ramEnd
    plo r2
    ;fall through to sep R3 which jumps to main

;; -Start of Standard Call and Return Technique;;  
;; sep R4 nnnn is CALL nnnn
;; sep R5 is RET
;; R6 points to address after CALL nnnn
;; R3 is program counter
;; R2 is stack

exita:
    sep r3	; 

call:
; set x to stack
    sex r2

; save r6 to stack
    ghi r6
    stxd
    glo r6
    stxd

; move r3 into r6, to save return address
    ghi r3
    phi r6
    glo r3
    plo r6

; load the address of subroutine
    lda r6
    phi r3
    lda r6
    plo r3

; exit and run subroutine
    br low exita

exitr:
    sep r3

return:
; set x to stack
    sex r2
; copy r6 into r3
    ghi r6
    phi r3
    glo r6
    plo r3
; increment stack pointer
    inc r2
; restore pointer from stack
    ldxa
    plo r6
    ldx
    phi r6
; head back to R3 progam counter
    br low exitr

;; -End of Standard Call and Return Technique;; 

;; test programs for floating point demo
main:
	sep r4
	dw load_d	;load float constant in D
	db   "+1130000-04"

	sep r4	;convert D to AC binary
	dw fpm_D2AC

	sep r4	;move AC to V1
	dw fpm_AC2V
	db 01

	sep r4
	dw load_d	;load const in D
	db   "+3550000-04"

	sep r4	;convert D to AC binary
	dw fpm_D2AC

	sep r4	;move V1 to OP
	dw fpm_V2OP
	db 01

	sep r4	;divide AC by OP -> AC
	dw fpm_div

	sep r4	;convert AC to D decimal
	dw fpm_AC2D

	sep r4	;convert D to ASCII values
	dw D2String
	db dtest	;ASCII address

	sep r4	;print string
	dw puts
	db dteststring ;string address

loop:
    br low loop


;string to print
dteststring:
	db "D: "
dtest: ;loaded with quotient result
     db 00, 00, 00, 00, 00, 00, 00, 00, 00, 00, 00
     db 0dh, 0ah, 00	;string ends with null

puts:		;print this string

    lda r6
    phi ra
    lda r6
    plo ra
    sex ra	;sets up RA index to point to print string

puts_full:	;implementation specific print

	bn1 low puts_full	;loop until EF1 set
	out 2				;out to printer, console, etc.
	ldx
	bnz low puts_full
	sep r5			;return on null terminator

D2String:	;convert D to ASCII values into caller's buffer
		;routine saves a register by not transfer-and-loop
		;is longer but slightly faster no incr test br loop
	lda r6
	phi rb
	lda r6
	plo rb	;RB -> output ASCII buffer

	ldi high D_sign
	phi ra
	ldi low D_sign
	plo ra	
	sex ra	;RA = address 0114H, start of Decimal Register

	ldxa
	str rb	;transfer sign which is "+" or "-"
	inc rb

	ldxa
	ani 0fH	;fetch BCD value 0 thru 9
	adi 30H	;toascii
	str rb	;transfer

	inc rb
	ldxa
	ani 0fH	;fetch BCD value 0 thru 9
	adi 30H
	str rb

	inc rb
	ldxa
	ani 0fH	;next digit
	adi 30H
	str rb

	inc rb
	ldxa
	ani 0fH
	adi 30H
	str rb

	inc rb
	ldxa
	ani 0fH
	adi 30H
	str rb

	inc rb
	ldxa
	ani 0fH
	adi 30H
	str rb

	inc rb
	ldxa
	ani 0fH
	adi 30H
	str rb

	inc rb
	ldxa		;fetch exponent sign '+' or '-'
	str rb

	inc rb
	ldxa		;fetch exponent first digit
	ani 0fH
	adi 30H
	str rb

	inc rb
	ldxa		;fetch exponent second digit
	ani 0fH
	adi 30H
	str rb
     sep r5	;return

load_d:	;load ASCII decimal float constant into D
		;loading address-down to exploit STXD instr
	ldi high D_exp1
	phi ra
	ldi low D_exp1
	plo ra	;011EH, end of Decimal register
	sex ra

	lda r6 ; de low exp
	stxd
	lda r6 ; de high exp
	stxd
	lda r6 ; de exp sign
	stxd

	lda r6 ; de mantissa b6
	stxd
	lda r6 ; de mantissa b5
	stxd
	lda r6 ; de mantissa b4
	stxd
	lda r6 ; de mantissa b3
	stxd
	lda r6 ; de mantissa b2
	stxd
	lda r6 ; de mantissa b1
	stxd
	lda r6 ; de mantissa b0
	stxd
	lda r6 ; de mantissa sign
	stxd
	sep r5	;return

	;tack FPM.bin at 0100H after this program
	ds	(0100H-1)- $
	db	00

	end
